// import style from './style.css';
require('./style.css');
import SwoopButton from './swoop-button.js';
import jwt_decode from "jwt-decode";

export default class {
  constructor(sid, config, cb = null) {
    this.user = {
      loggedIn: false
    };
    this.oauth = {};
    this.config = {
      sid: sid,
      endpoint: 'https://app.swoop.email',
      apiEndpoint: 'https://api.swoop.email',
      platform: 'web',
      version: '1.4.12',
      intialized: false,
      redirectURI: window.location.href,
      onUserMeta: () => {},
      onShow: () => {},
      onSuccess: () => {},
      onLoading: () => {},
      onError: () => {},
      onClose: () => {},
      ...config
    };
    this._loadCached();
  }

  configure(config = {}) {
    config.sid = this.config.sid || config.swoopClientId || config.sid;
    this.config = {
      ...this.config,
      ...config,
    };
  }

  button() {
    // Create swoop buttons
    const swoopButtons = document.querySelectorAll(".swoop-button");
    let config = {};
    if (window.swoop.user && window.swoop.user.email) {
      config.email = window.swoop.user.email;
    }
    for (let i = 0; i < swoopButtons.length; i++) {
      let button = swoopButtons[i];
      new SwoopButton(button, config);
    }
  }

  async checkForSwoopOauth2Code(code) {
    return new Promise(async(resolve, reject) => {
      if (code && !this.user.loggedIn) {
        if (this.config.onLoading) {
          this.config.onLoading();
        }
        // exchange the code for a token
        let response;
        try {
          response = await this._exchangeCodeForToken(code);
        } catch (e) {
          return reject(e);
        }
        this.oauth = response;
        this._hydrateUserFromToken(response.id_token);
        window.localStorage.setItem("swoop", JSON.stringify(this.oauth));
        if (this.config.onSuccess) {
          this.config.onSuccess(this.user);
        }

        if (this.w) {
          this.w.close();
        }

        resolve(this.user);
      } else {
        resolve(null);
      }
    });
  }


  async in (userMeta = null, appMeta = null) {

    if (this.config.onUserMeta) {
      userMeta = {
        ...userMeta,
        ...this.config.onUserMeta()
      }
    }

    let defaultMode = this.config.mode || 'popup';
    let w = null;
    let url = await this._buildSwoopOauthWithPkceUrl(defaultMode, userMeta, appMeta);

    if (defaultMode == 'popup') {
      let width = 500;
      let height = 600;
      var left = (screen.width - width) / 2;
      var top = (screen.height - height) / 4;
      let params = `scrollbars=no,resizable=yes,status=no,location=no,toolbar=no,menubar=no,
  width=${width},height=${height},left=${left},top=${top}`;
      w = open(url, '_blank', params);

      if (this.config.onShow) {
        this.config.onShow(w);
      }
    }

    if (defaultMode === 'redirect' || !w || w.closed || typeof w.closed == 'undefined') {
      url = await this._buildSwoopOauthWithPkceUrl('redirect', userMeta, appMeta);
      window.location.href = url;
    } else {
      // Start a timeout to check for closed window
      let timer = setInterval(() => {
        if (w.closed) {
          if (this.config.onClose) {
            this.config.onClose();
          }
          clearInterval(timer);
        }
      }, 5000);
    }
    this.w = w;
  }

  out() {
    // Remove cached user
    window.localStorage.removeItem("swoop");
    this.user = {};
  }

  // OAuth call to exchange a code for a token
  async _exchangeCodeForToken(code) {
    return new Promise((resolve, reject) => {
      var xhr = new XMLHttpRequest();
      xhr.onload = function() {
        var response = xhr.response;
        var message;
        if (xhr.status == 200) {
          resolve(response)
        } else {
          reject(new Error('Unable to fetch id token'))
        }
      };
      xhr.responseType = 'json';
      xhr.open("POST", `${this.config.apiEndpoint}/oauth2/token`, true);
      xhr.setRequestHeader('Content-type', 'application/x-www-form-urlencoded');
      xhr.send(new URLSearchParams({
        client_id: this.config.sid,
        code_verifier: window.localStorage.getItem("code_verifier"),
        grant_type: "authorization_code",
        redirect_uri: this.config.redirectURI,
        code: code
      }));
    })
  }

  // Creates the OAuth URL for Swoop
  // Mode can be popup or redirect
  _buildSwoopOauthWithPkceUrl(mode = 'redirect', userMeta = null, appMeta = null) {
    return new Promise((resolve, reject) => {
      let codeVerifier = this._generateRandomString(64);
      this._pkceChallengeFromVerifier(codeVerifier)
        .then(codeChallenge => {
          window.localStorage.setItem("code_verifier", codeVerifier);

          let urlParts = window.location.href.split('?');
          let redirectUri = urlParts[0];

          let args = new URLSearchParams({
            scope: "email",
            response_type: "code",
            client_id: this.config.sid,
            code_challenge_method: "S256",
            code_challenge: codeChallenge,
            redirect_uri: this.config.redirectURI,
            target: 'swoop.js',
            platform: this.config.platform,
            version: this.config.version,
            mode
          });

          // Append the user meta params
          if (userMeta) {
            let userMetaParams = "";
            for (let umKey in userMeta) {
              let val = userMeta[umKey];
              args.set(`user_meta[${encodeURIComponent(umKey)}]`, encodeURIComponent(val));
            };
          }

          // Append the app meta params
          if (appMeta) {
            let appMetaParams = "";
            for (let amKey in appMeta) {
              let val = appMeta[amKey];
              args.set(`app_meta[${encodeURIComponent(amKey)}]`, encodeURIComponent(val));
            };
          }

          resolve(`${this.config.endpoint}/oauth2/authorize?${args}`);
        })
        .catch(e => {
          reject(e);
        })
    })
  }

  // Return the base64-urlencoded sha256 hash for the PKCE challenge
  _pkceChallengeFromVerifier(v) {
    return new Promise((resolve, reject) => {
      this._sha256(v)
        .then(hashed => {
          resolve(this._base64urlencode(hashed));
        })
    });
  }

  // Calculate the SHA256 hash of the input text.
  // Returns a promise that resolves to an ArrayBuffer
  _sha256(plain) {
    const encoder = new TextEncoder();
    const data = encoder.encode(plain);
    return window.crypto.subtle.digest('SHA-256', data);
  }

  // Base64-urlencodes the input string
  _base64urlencode(str) {
    // Convert the ArrayBuffer to string using Uint8 array to conver to what btoa accepts.
    // btoa accepts chars only within ascii 0-255 and base64 encodes them.
    // Then convert the base64 encoded to base64url encoded
    //   (replace + with -, replace / with _, trim trailing =)
    return btoa(String.fromCharCode.apply(null, new Uint8Array(str)))
      .replace(/\+/g, '-').replace(/\//g, '_').replace(/=+$/, '');
  }

  _generateRandomString(length) {
    var text = "";
    var possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";

    for (var i = 0; i < length; i++) {
      text += possible.charAt(Math.floor(Math.random() * possible.length));
    }

    return text;
  }

  _loadCached() {
    // Check cached user
    let cached = window.localStorage.getItem("swoop");
    if (cached && cached != null && typeof cached === 'string') {
      let oauth;
      try {
        // Parse the cached swoop payload
        oauth = JSON.parse(cached);
        // Attempt to hydrate this user (if their token is still valid)
        // TODO: Check refresh token and fetch new ones
        if (!this._hydrateUserFromToken(oauth.id_token)) {
          window.localStorage.removeItem("swoop");
        } else {
          this.oauth = oauth;
        }
      } catch (e) {
        // do nothing.
      }
    }
  }

  _hydrateUserFromToken(token) {
    let decoded = jwt_decode(token);

    // Check expiration
    let d = new Date();
    let n = d.getTime() / 1000;

    if (decoded.exp - n < 0) {
      return false;
    }

    this.user.email = decoded.email;
    this.user.id_token = token;
    this.user.uid = decoded.sub;
    if (decoded.user_meta) {
      this.user.user_meta = this._deepUrlDecode(decoded.user_meta);
    }

    return true;
  }

  _deepUrlDecode(obj) {
    if (typeof obj === 'string') {
      return decodeURIComponent(obj);
    } else if (typeof obj === 'object') {
      for (let key in obj) {
        obj[key] = this._deepUrlDecode(obj[key]);
      }
    }
    return obj;
  }

}